"""
Script to generate test data for fabulous-paths!

Usage:
    python generate_test_data.py

Creates a file called test_data.db with the brief tests (fake TPS with
6 steps of one-way shooting in a fixed length 5-frame ensemble -- all paths
will be accepted).

This file is used in the test suite.
"""

import mdtraj as md
import openmmtools
import openpathsampling as paths


### SIMSTORE SETUP #########################################################
from openpathsampling.experimental.storage import Storage, monkey_patch_all
from openpathsampling.experimental.storage.collective_variables import \
        MDTrajFunctionCV
paths = monkey_patch_all(paths)
# don't need to worry about interface sets here


### MAIN SETUP #############################################################
print("Setting up the simulation")
testsystem = openmmtools.testsystems.AlanineDipeptideVacuum()
integrator = openmmtools.integrators.VVVRIntegrator()
topology = paths.engines.openmm.topology.MDTrajTopology(testsystem.mdtraj_topology)

engine = paths.engines.openmm.Engine(topology=topology,
                                     system=testsystem.system,
                                     integrator=integrator,
                                     options={
                                         'n_frames_max': 100,
                                         'n_steps_per_frame': 10,
                                     }).named('fake_engine')

ensemble=paths.LengthEnsemble(5).named("length 5")
mover = paths.OneWayShootingMover(ensemble=ensemble,
                                  selector=paths.UniformSelector(),
                                  engine=engine)

scheme = paths.LockedMoveScheme(root_mover=mover)


### INITIAL SNAPSHOT #######################################################
print("Getting energy-minimized initial snapshot")
engine.simulation.context.setPositions(testsystem.positions)
engine.simulation.minimizeEnergy()
initial_snapshot = engine.current_snapshot


### SETUP INITIAL CONDITIONS ###############################################
print("Setting up initial conditions")
initial_trajectory = engine.generate(initial_snapshot, ensemble.can_append)
sample = paths.Sample(replica=0,
                      ensemble=ensemble,
                      trajectory=initial_trajectory)
initial_conditions = paths.SampleSet([sample])

ref_frame = initial_trajectory.to_mdtraj()[0]
ref_frame.save("ref.pdb")

### RUN PATH SAMPLING ######################################################
storage = Storage('test_data.db', mode='w')
# add CVs
phi = MDTrajFunctionCV(md.compute_dihedrals,
                       topology=engine.topology,
                       indices=[[4, 6, 8, 14]]).named("phi")

psi = MDTrajFunctionCV(md.compute_dihedrals,
                       topology=engine.topology,
                       indices=[[6, 8, 14, 16]]).named("psi")

storage.save([phi, psi])

sim = paths.PathSampling(
    storage=storage,
    move_scheme=scheme,
    sample_set=initial_conditions
)
sim.run(6)
storage.close()



from tqdm.auto import tqdm
import pandas as pd
import numpy as np

def extract_CV(traj, cvs):
    """Create a CV dataframe for FABULOUS from an OPS trajectory and CVs.

    Returns
    -------
    pandas.DataFrame :
        each row is a snapshot
    """
    columns = [cv.name for cv in cvs]
    results = np.array([cv(traj) for cv in cvs]).T
    return pd.DataFrame(results, columns=columns)


def extract_MD(trajectory, ref, keep_atoms):
    """Create an MD dataframe for FABULOUS from an OPS trajectory.

    Parameters
    ----------
    trajectory : :class:`openpathsampling.Trajectory`
    ref : :class:`mdtraj.Trajectory`
    keep_atoms : List[int]

    Returns
    -------
    pandas.DataFrame :
        each row is a snapshot, all coordinates in the columns
    """
    traj = trajectory.to_mdtraj().atom_slice(keep_atoms)
    traj.superpose(ref.atom_slice(keep_atoms))  # TODO: move this out

    xyz = traj.xyz.reshape(traj.n_frames, traj.n_atoms * 3)
    return pd.DataFrame(xyz)


def _get_keep_atoms(topology, keep):
    if isinstance(keep, str):
        return topology.select(keep)
    else:
        return [int(i) for i in keep]


def extract_OPS(steps, ref, keep, cvs):
    """Full process of extracting CV and trajectory dataframes fom OPS.

    Parameters
    ----------
    steps: List[:class:`openpathsampling.MCStep`]
        input steps
    """
    from openpathsampling.analysis.tis.core import \
        steps_to_weighted_trajectories
    ensembles = [steps[0].active[0].ensemble]  # TODO: remove later
    keep_atoms = None
    desc1 = "Identifying trajectories"
    weighted_trajs = steps_to_weighted_trajectories(tqdm(steps, desc=desc1),
                                                    ensembles)
    if len(weighted_trajs) != 1:
        # NOTE: this is currently impossible, but might become possible when
        # OPS fixes steps_to_weighted to not require ensembles as input
        raise RuntimeError("FABULOUS requires all steps from the same",
                           f"ensemble. Found {len(weighted_trajs)} ",
                           "ensembles.")

    traj_counter = list(weighted_trajs.values())[0]

    desc2 = "Extracting data for FABULOUS"
    for traj in tqdm(traj_counter, desc=desc2):
        if keep_atoms is None:
            # this only runs on the first trajectory
            topology = traj[0].topology.mdtraj
            keep_atoms = _get_keep_atoms(topology, keep)

        trj_df = extract_MD(traj, ref, keep_atoms)
        cv_df = extract_CV(traj, cvs)
        yield trj_df, cv_df

def _concat_dfs(dfs):
    traj_dfs, cv_dfs = zip(*dfs)
    trajs = pd.concat(traj_dfs, ignore_index=True)
    cvs = pd.concat(cv_dfs, ignore_index=True)
    return trajs, cvs

def main(steps, cvs, ref, keep, yaml_file, n_gen, results_dir, label):
    dfs = extract_OPS(steps, ref, keep, cvs)
    trajs, cvs = _concat_dfs(dfs)
    print("Trajectories:")
    print(trajs.head())
    print("CVs:")
    print(cvs.head())
    ... # from here there should be something in FABULOUS that does the work
    pass



import pytest
from unittest import mock
from openpathsampling.tests.test_helpers import make_1d_traj
from openpathsampling.analysis.tis.core import \
    steps_to_weighted_trajectories
import pandas as pd
import numpy as np
import os

import pkg_resources

from fabulous_paths import *
from fabulous_paths import _get_keep_atoms, _concat_dfs

BACKBONE = [4, 5, 6, 8, 14, 15, 16, 18]

@pytest.fixture(scope="session")
def datafile():
    import openpathsampling as paths
    from openpathsampling.experimental import storage
    paths = storage.monkey_patch_all(paths)
    filename = pkg_resources.resource_filename(
        'fabulous_paths',
        os.path.join('tests', 'test_data', 'test_data.db')
    )
    if not os.path.exists(filename):  # pragma: no cover
        pytest.skip("Test file missing")
    datafile = storage.Storage(filename, mode='r')
    return datafile

@pytest.fixture
def trajectory_data(datafile):
    ensemble = datafile.ensembles['length 5']
    weighted_trajs = steps_to_weighted_trajectories(datafile.steps,
                                                    ensembles=[ensemble])
    traj_counter = list(weighted_trajs.values())[0]
    # we run 6 trajectories, plus initial
    assert sum(traj_counter.values()) == 7
    trajectories = list(traj_counter.keys())
    for traj in trajectories:
        assert len(traj) == 5  # each trajectory is 5 frames
    return trajectories

def test_extract_CV(datafile, trajectory_data):
    phi = datafile.cvs['phi']
    psi = datafile.cvs['psi']
    cvs = [phi, psi]
    for traj in trajectory_data:
        extracted = extract_CV(traj, cvs)
        assert extracted.shape == (len(traj), len(cvs))
        assert all(extracted.columns == ['phi', 'psi'])

def test_extract_MD(trajectory_data):
    for traj in trajectory_data:
        # use the first frame as ref; gives us a no-change to test
        ref = traj.to_mdtraj()
        extracted = extract_MD(traj, ref[0], BACKBONE)
        n_cols_expected = len(BACKBONE) * 3
        assert extracted.shape == (len(traj), n_cols_expected)
        first_frame = extracted.loc[0,:]
        sliced_ref = ref.atom_slice(BACKBONE)
        np.testing.assert_allclose(first_frame, sliced_ref.xyz[0].flatten(),
                                   rtol=1e-6)  # seems we need extra room

@pytest.mark.parametrize('type_', ['str', 'list'])
def test_get_keep_atoms(trajectory_data, type_):
    keep = {'str': 'backbone', 'list': BACKBONE}[type_]
    topology = trajectory_data[0][0].topology.mdtraj
    np.testing.assert_array_equal(_get_keep_atoms(topology, keep), BACKBONE)

@mock.patch('fabulous_paths.tqdm', new=lambda x, desc: x)
def test_extract_OPS(datafile):
    ref = datafile.steps[0].active[0].trajectory.to_mdtraj()[0]
    keep = BACKBONE
    cvs = [datafile.cvs['phi'], datafile.cvs['psi']]
    dfs = list(extract_OPS(datafile.steps, ref, keep, cvs))
    assert len(dfs) == len(datafile.steps)
    for trj_df, cv_df in dfs:
        # 5 == len(traj)
        # 2 == len(cvs)
        assert trj_df.shape == (5, len(BACKBONE) * 3)
        assert cv_df.shape == (5, 2)

def test_extract_OPS_bad_ensembles():
    # this can't be tested yet, because it is (temporarily) impossible
    pytest.skip()

def test_concat_dfs():
    sizes = [(4, 2), (4, 3)]
    dfs = [tuple(pd.DataFrame(np.random.random(size)) for size in sizes)
           for _ in range(2)]
    # sanity check
    for df_pair in dfs:
        for df, shape in zip(df_pair, sizes):
            assert df.shape == shape

    trajs, cvs = _concat_dfs(dfs)
    assert trajs.shape == (8, 2)
    assert cvs.shape == (8, 3)

def test_main_integration(datafile, tmpdir):
    # smoke test to ensure that the integration with FABULOUS works
    ref = datafile.steps[0].active[0].trajectory.to_mdtraj()[0]
    keep = BACKBONE
    cvs = [datafile.cvs['phi'], datafile.cvs['psi']]
    results_dir=tmpdir / "results"
    label='1'
    yaml_file=... # TODO
    main(steps=datafile.steps, cvs=cvs, ref=ref, keep=keep,
         yaml_file=yaml_file, n_gen=0, results_dir=results_dir, label=label)



import os
import click
from fabulous_paths import main
import mdtraj as md
from paths_cli.parameters import INPUT_FILE, MULTI_CV

@click.command(
    "fabulous",
    short_help="Analysis using the FABULOUS framework"
)
@INPUT_FILE.clicked(required=True)
@MULTI_CV.clicked(required=True)   # TODO: option to allow --cv-list file?
@click.option('--ref', type=click.Path(readable=True), required=True,
              help="reference frame for alignment")
@click.option('--keep-atoms', type=str, required=True,
              help="atoms to keep")  # TODO: also allow `.npy` files
@click.option('-c', '--conf', type=click.Path(readable=True), required=True,
              help="FABULOUS configuration file")
@click.option('-n', '--ngen', type=int,
              help="number of generations to run")
@click.option('--results', type=click.Path(writable=True), required=True,
              help="directory to store results in")
@click.option('--label', type=str, required=False, default="1",
              help="label for this run")
def fabulous(input_file, cv, ref, keep_atoms, conf, ngen, results, label):
    storage = INPUT_FILE.get(input_file)
    cvs = MULTI_CV.get(storage, cv)
    ref = md.load(ref)
    main(storage.steps, cvs, ref, keep_atoms, conf, ngen, results, label)


# _mock_command used in docs generation
_mock_command = click.Group(name='openpathsampling',
                            commands={'fabulous': fabulous})


CLI = fabulous
SECTION = "Analysis"
OPS_VERSION = (1, 5)
